#include "stdafx.h"
#include <Mmsystem.h>
#include "CPUInfo.h"

#ifndef __cpuid
void __cpuid(uint32 CPUInfo[4], uint32 InfoType)
{
	__asm {
		mov eax, InfoType
		cpuid
		mov esi, CPUInfo
		mov [esi], eax
		mov [esi+4], ebx
		mov [esi+8], ecx
		mov [esi+12], edx
	}
}
#endif

#ifndef __cpuidex
void __cpuidex(uint32 CPUInfo[4], uint32 InfoType, uint32 ECXValue)
{
	__asm {
		mov eax, InfoType
		mov ecx, ECXValue
		cpuid
		mov esi, CPUInfo
		mov [esi], eax
		mov [esi+4], ebx
		mov [esi+8], ecx
		mov [esi+12], edx
	}
}
#endif

// borrowed from pcsx2 emulator
uint64 GetCPUSpeedHz( unsigned int time )
{
	sint64 timeStart, timeStop;
	sint64 startTick, endTick;
	sint64 overhead;

	overhead = __rdtsc() - __rdtsc();

	timeStart = timeGetTime( );
	while( timeGetTime( ) == timeStart ) 
	{
		timeStart = timeGetTime( );
	}
	for(;;)
	{
		timeStop = timeGetTime( );
		if ( ( timeStop - timeStart ) > 1 )	
		{
			startTick = __rdtsc( );
			break;
		}
	}

	timeStart = timeStop;
	for(;;)
	{
		timeStop = timeGetTime( );
		if ( ( timeStop - timeStart ) > time )	
		{
			endTick = __rdtsc( );
			break;
		}
	}

	return (uint64)( ( endTick - startTick ) + ( overhead ) );
}

// borrowed from MSDN2008
void GetCpuInfo(CPUINFO *cpu) 
{
	uint32 CPUInfo[4];

	if(cpu == NULL) return;
	memset(cpu, 0, sizeof(CPUINFO));
	
	// load Vendor String
	__cpuid(CPUInfo, 0);
	((uint32*)cpu->szID)[0] = CPUInfo[1];
	((uint32*)cpu->szID)[1] = CPUInfo[3];
	((uint32*)cpu->szID)[2] = CPUInfo[2];

	// load code 1
	__cpuid(CPUInfo, 1);
	cpu->nSteppingID = CPUInfo[0] & 0xf;
	cpu->nModel = (CPUInfo[0] >> 4) & 0xf;
	cpu->nFamily = (CPUInfo[0] >> 8) & 0xf;
	cpu->nProcessorType = (CPUInfo[0] >> 12) & 0x3;
	cpu->nExtendedModel = (CPUInfo[0] >> 16) & 0xf;
	cpu->nExtendedFamily = (CPUInfo[0] >> 20) & 0xff;
	cpu->nBrandIndex = CPUInfo[1] & 0xff;
	cpu->nCLFLUSHCacheLineSize = ((CPUInfo[1] >> 8) & 0xff) * 8; // why?
	cpu->nLogicalProcessors = ((CPUInfo[1] >> 16) & 0xff);
	cpu->nAPICPhysicalID = (CPUInfo[1] >> 24) & 0xff;

	cpu->hasSSE3 = (CPUInfo[2] & 0x1) || false;
	cpu->hasMONITOR_MWAIT = (CPUInfo[2] & 0x8) || false;
	cpu->hasCPLQualifiedDebugStore = (CPUInfo[2] & 0x10) || false;
	cpu->hasVirtualMachineExtensions = (CPUInfo[2] & 0x20) || false;
	cpu->hasEnhancedIntelSpeedStepTechnology = (CPUInfo[2] & 0x80) || false;
	cpu->hasThermalMonitor2 = (CPUInfo[2] & 0x100) || false;
	cpu->hasSSSE3 = (CPUInfo[2] & 0x200) || false;
	cpu->hasL1ContextID = (CPUInfo[2] & 0x400) || false;
	cpu->hasCMPXCHG16B= (CPUInfo[2] & 0x2000) || false;
	cpu->hasxTPRUpdateControl = (CPUInfo[2] & 0x4000) || false;
	cpu->hasPerfDebugCapabilityMSR = (CPUInfo[2] & 0x8000) || false;
	cpu->hasSSE41 = (CPUInfo[2] & 0x80000) || false;
	cpu->hasSSE42 = (CPUInfo[2] & 0x100000) || false;
	cpu->hasPOPCNT = (CPUInfo[2] & 0x800000) || false;

	cpu->hasx87FPU = (CPUInfo[3] & 0x1) || false;
	cpu->hasVirtual8086ModeEnhancement = (CPUInfo[3] & 0x2) || false;
	cpu->hasDebugExt = (CPUInfo[3] & 0x4) || false;
	cpu->hasPageSizeExt = (CPUInfo[3] & 0x8) || false;
	cpu->hasTimeStampCounter = (CPUInfo[3] & 0x10) || false;
	cpu->hasRDMSR_WRMSRSupport = (CPUInfo[3] & 0x20) || false;
	cpu->hasPAE = (CPUInfo[3] & 0x40) || false;
	cpu->hasMachineCheckException = (CPUInfo[3] & 0x80) || false;
	cpu->hasCMPXCHG8B = (CPUInfo[3] & 0x100) || false;
	cpu->hasAPIC = (CPUInfo[3] & 0x200) || false;
	//Unknown1
	cpu->hasSYSENTER_SYSEXIT = (CPUInfo[3] & 0x800) || false;
	cpu->hasMTRRegs = (CPUInfo[3] & 0x1000) || false;
	cpu->hasPTEGlobalBit = (CPUInfo[3] & 0x2000) || false;
	cpu->hasMachineCheckArchitecture = (CPUInfo[3] & 0x4000) || false;
	cpu->hasCondMoveCompareInstruction = (CPUInfo[3] & 0x8000) || false;
	cpu->hasPageAttributeTable = (CPUInfo[3] & 0x10000) || false;
	cpu->has36bitPageSizeExt = (CPUInfo[3] & 0x20000) || false;
	cpu->hasSerialNumber = (CPUInfo[3] & 0x40000) || false;
	cpu->hasCFLUSHExt = (CPUInfo[3] & 0x80000) || false;
	//Unknown2
	cpu->hasDebugStore = (CPUInfo[3] & 0x200000) || false;
	cpu->hasThermalMonitorAndClockCtrl = (CPUInfo[3] & 0x400000) || false;
	cpu->hasMMX = (CPUInfo[3] & 0x800000) || false;
	cpu->hasFXSAVE_FXRSTOR = (CPUInfo[3] & 0x1000000) || false;
	cpu->hasSSE = (CPUInfo[3] & 0x2000000) || false;
	cpu->hasSSE2 = (CPUInfo[3] & 0x4000000) || false;
	cpu->hasSelfSnoop = (CPUInfo[3] & 0x8000000) || false;
	cpu->hasMultithreading = (CPUInfo[3] & 0x10000000) || false;
	cpu->hasThermalMonitor = (CPUInfo[3] & 0x20000000) || false;
	//Unknown4
	cpu->hasPendingBreakEnable = (CPUInfo[3] & 0x80000000) || false;

	//load code 0x80000001
	__cpuid(CPUInfo, 0x80000001);
	cpu->hasLAHF_SAHFAvailable = (CPUInfo[2] & 0x1) || false;
	cpu->hasCmpLegacy = (CPUInfo[2] & 0x2) || false;
	cpu->hasSVM = (CPUInfo[2] & 0x4) || false;
	cpu->hasExtApicSpace = (CPUInfo[2] & 0x8) || false;
	cpu->hasAltMovCr8 = (CPUInfo[2] & 0x10) || false;
	cpu->hasLZCNT = (CPUInfo[2] & 0x20) || false;
	cpu->hasSSE4A = (CPUInfo[2] & 0x40) || false;
	cpu->hasMisalignedSSE = (CPUInfo[2] & 0x80) || false;
	cpu->hasPREFETCH = (CPUInfo[2] & 0x100) || false;
	cpu->hasSKINITandDEV = (CPUInfo[2] & 0x1000) || false;
	cpu->hasSYSCALL_SYSRETAvailable = (CPUInfo[3] & 0x800) || false;
	cpu->hasExecuteDisableBitAvailable = (CPUInfo[3] & 0x10000) || false;
	cpu->hasMMXExt = (CPUInfo[3] & 0x400000) || false;
	cpu->hasFFXSR = (CPUInfo[3] & 0x2000000) || false;
	cpu->has1GBSupport = (CPUInfo[3] & 0x4000000) || false;
	cpu->hasRDTSCP = (CPUInfo[3] & 0x8000000) || false; //bit27
	cpu->has64Available = (CPUInfo[3] & 0x20000000) || false; //bit29
	cpu->has3DNowExt = (CPUInfo[3] & 0x40000000) || false; //bit30
	cpu->has3DNow = (CPUInfo[3] & 0x80000000) || false; //bit31

	// Interpret CPU brand string.
	__cpuid(CPUInfo, 0x80000002);
	memcpy(cpu->szBrandString, CPUInfo, sizeof(CPUInfo));
	__cpuid(CPUInfo, 0x80000003);
	memcpy(cpu->szBrandString + 16, CPUInfo, sizeof(CPUInfo));
	__cpuid(CPUInfo, 0x80000004);
	memcpy(cpu->szBrandString + 32, CPUInfo, sizeof(CPUInfo));

	__cpuid(CPUInfo, 0x80000006);
	cpu->nCacheLineSize = CPUInfo[2] & 0xff;
	cpu->nL2Associativity = (CPUInfo[2] >> 12) & 0xf;
	cpu->nCacheSizeK = (CPUInfo[2] >> 16) & 0xffff;

	__cpuid(CPUInfo, 0x80000008);
	cpu->nPhysicalAddress = CPUInfo[0] & 0xff;
	cpu->nVirtualAddress = (CPUInfo[0] >> 8) & 0xff;

	__cpuid(CPUInfo, 0x8000000A);
	cpu->hasNestedPaging = (CPUInfo[3] & 0x1) || false;
	cpu->hasLBRVisualization = (CPUInfo[3] & 0x2) || false;

	__cpuid(CPUInfo, 0x8000001A);
	cpu->hasFP128 = (CPUInfo[0] & 0x1) || false;
	cpu->hasMOVOptimization = (CPUInfo[0] & 0x2) || false;

	__cpuidex(CPUInfo, 0x4, 0);
	cpu->nCores = (CPUInfo[0] >> 26) + 1;

	if(cpu->hasTimeStampCounter)
	{
		cpu->nCPUSpeed = (uint32 )(GetCPUSpeedHz( 1000 ) / 1000000);
	}
}