#pragma once

#ifndef __CPUINFO_H__
#define __CPUINFO_H__

typedef struct {
	uint64 nCPUSpeed;                           // in GHz

	sint8  szID[16];                            // Vendor ID (%s)
	sint8  szBrandString[64];                   // Brand String (%s)

	uint8  nSteppingID;                         // Stepping ID (%d)
	uint8  nModel;                              // Model (%d) 
	uint8  nFamily;                             // Family (%d)
	uint8  nProcessorType;                      // Processor Type (%d)
	uint8  nExtendedModel;                      // Extended Model (%d)
	uint8  nExtendedFamily;                     // Extended Family (%d)
	uint8  nBrandIndex;                         // nBrandIndex (%d)
	uint8  nCLFLUSHCacheLineSize;               // CLFLUSH cache line size (%d)
	uint8  nLogicalProcessors;                  // Logical Processor Count (%d)
	uint8  nAPICPhysicalID;                     // APIC Physical ID (%d)
	uint8  nCores;                              // Number of Cores (%d)

	uint8  nCacheLineSize;                      // Cache Line Size
	uint8  nL2Associativity;                    // L2 Associativity
	uint32 nCacheSizeK;                         // Cache Size (in KBytes)
	uint8  nPhysicalAddress;                    // Physical address bits 
	uint8  nVirtualAddress;                     // Virtual address bits 
	
	bool   hasMMX;                              // MMX Extensions
	bool   hasMMXExt;                           // Extensions to MMX Instructions
	bool   has3DNow;                            // 3Dnow! instructions
	bool   has3DNowExt;                         // 3Dnow! Ext
	bool   hasSSE;                              // SSE Extensions
	bool   hasSSE2;                             // SSE2 Extensions
	bool   hasSSE3;                             // SSE3 Extensions
	bool   hasSSSE3;                            // Supplemental SSE3 Extensions
	bool   hasSSE41;                            // SSE4.1 Extensions
	bool   hasSSE42;                            // SSE4.2 Extensions
	bool   hasSSE4A;                            // SSE4A (EXTRQ, INSERTQ, MOVNTSD, MOVNTSS)
	bool   hasMONITOR_MWAIT;                    // MONITOR/MWAIT
	bool   hasCPLQualifiedDebugStore;           // CPL Qualified Debug Store
	bool   hasVirtualMachineExtensions;         // Virtual Machine Extensions
	bool   hasEnhancedIntelSpeedStepTechnology; // Enhanced Intel SpeedStep Technology
	bool   hasThermalMonitor2;                  // Thermal Monitor 2
	bool   hasL1ContextID;                      // L1 Context ID
	bool   hasCMPXCHG16B;                       // CMPXCHG16B Instruction
	bool   hasxTPRUpdateControl;                // xTPR Update Control
	bool   hasPerfDebugCapabilityMSR;           // Perf\Debug Capability MSR
	bool   hasPOPCNT;                           // PPOPCNT Instruction
	bool   hasLAHF_SAHFAvailable;               // LAHF/SAHF in 64-bit mode
	bool   hasCmpLegacy;                        // Core multi-processing legacy mode
	bool   hasSVM;                              // Secure Virtual Machine
	bool   hasExtApicSpace;                     // Extended APIC Register Space
	bool   hasAltMovCr8;                        // AltMovCr8
	bool   hasLZCNT;                            // LZCNT instruction
	bool   hasMisalignedSSE;                    // Misaligned SSE mode
	bool   hasPREFETCH;                         // PREFETCH and PREFETCHW Instructions
	bool   hasSKINITandDEV;                     // SKINIT and DEV support
	bool   hasSYSCALL_SYSRETAvailable;          // SYSCALL/SYSRET in 64-bit mode
	bool   hasExecuteDisableBitAvailable;       // Execute Disable Bit

	bool   hasFFXSR;                            // FXSR
	bool   has1GBSupport;                       // 1GB page support
	bool   hasRDTSCP;                           // RDTSCP instruction
	bool   has64Available;                      // 64 bit Technology
	bool   hasNestedPaging;                     // Nested Paging
	bool   hasLBRVisualization;                 // LBR Visualization
	bool   hasFP128;                            // FP128 optimization
	bool   hasMOVOptimization;                  // MOVU Optimization

	bool   hasx87FPU;                           // x87 FPU On Chip
	bool   hasVirtual8086ModeEnhancement;       // Virtual-8086 Mode Enhancement
	bool   hasDebugExt;                         // Debugging Extensions
	bool   hasPageSizeExt;                      // Page Size Extensions
	bool   hasTimeStampCounter;                 // Time Stamp Counter
	bool   hasRDMSR_WRMSRSupport;               // RDMSR and WRMSR Support
	bool   hasPAE;                              // Physical Address Extensions
	bool   hasMachineCheckException;            // Machine Check Exception
	bool   hasCMPXCHG8B;                        // CMPXCHG8B Instruction
	bool   hasAPIC;                             // APIC On Chip
	bool   hasSYSENTER_SYSEXIT;                 // SYSENTER and SYSEXIT
	bool   hasMTRRegs;                          // Memory Type Range Registers
	bool   hasPTEGlobalBit;                     // PTE Global Bit
	bool   hasMachineCheckArchitecture;         // Machine Check Architecture
	bool   hasCondMoveCompareInstruction;       // Conditional Move/Compare Instruction
	bool   hasPageAttributeTable;               // Page Attribute Table
	bool   has36bitPageSizeExt;                 // 36-bit Page Size Extension
	bool   hasSerialNumber;                     // Processor Serial Number
	bool   hasCFLUSHExt;                        // CFLUSH Extension
	bool   hasDebugStore;                       // Debug Store
	bool   hasThermalMonitorAndClockCtrl;       // Thermal Monitor and Clock Ctrl
	bool   hasFXSAVE_FXRSTOR;                   // FXSAVE/FXRSTOR
	bool   hasSelfSnoop;                        // Self Snoop
	bool   hasMultithreading;                   // Multithreading Technology
	bool   hasThermalMonitor;                   // Thermal Monitor
	bool   hasPendingBreakEnable;               // Pending Break Enable
} CPUINFO;

void GetCpuInfo(CPUINFO *cpuinfo);

#endif // __CPUINFO_H__
