//this file is part of eMule IS Mod
//Copyright (C)2009 eMule IS Mod
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <atlutil.h>
#include <IPHlpApi.h>
#include <openssl/err.h>
#include <openssl/conf.h>
#include "tools.h"
#include "MD5Sum.h"
#include "options.h"
#include "sctCore.h"
#include "SecureSocket.h"
#include "resource.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

typedef HANDLE (__stdcall *PROC_IcmpCreateFile)(void); 
PROC_IcmpCreateFile IcmpCreateFileDyn;

typedef DWORD (__stdcall *PROC_IcmpSendEcho)(HANDLE,IPAddr,LPVOID,WORD,PIP_OPTION_INFORMATION,LPVOID,DWORD,DWORD);
PROC_IcmpSendEcho IcmpSendEchoDyn;

typedef BOOL (__stdcall *PROC_IcmpCloseHandle)(HANDLE);
PROC_IcmpCloseHandle IcmpCloseHandleDyn;

RSA* LoadPublicKey();

// 
CSctCore::CSctCore()
    : m_bAlreadyUpdated(false),
      m_uDelay(SCT_FIRST_DELAY),
      m_strLogin(0),
      m_strPassword(0),
      m_clientThread(0),
      m_serverThread(0)
{
    // load icmp functions
	m_hIcmpDll = LoadLibrary(_T("iphlpapi.dll"));
    if ( m_hIcmpDll == NULL ) // xp and later
    {
        m_hIcmpDll = LoadLibrary(_T("icmp.dll"));
    }
    IcmpCreateFileDyn = (PROC_IcmpCreateFile)GetProcAddress(m_hIcmpDll, "IcmpCreateFile");
    IcmpSendEchoDyn = (PROC_IcmpSendEcho)GetProcAddress(m_hIcmpDll, "IcmpSendEcho");
    IcmpCloseHandleDyn = (PROC_IcmpCloseHandle)GetProcAddress(m_hIcmpDll, "IcmpCloseHandle");

    LoadSettings();
    SecureSocket::loadSsl();
    m_hWaitEvent = CreateEvent(0,0,0,0);
    InitializeCriticalSection(&m_updateServersCS);

    //   
    CStringVector arrIf = GetChildRegKeys(SCT_INTERFACES_PATH);
    for( UINT i = 0; i < arrIf.size(); ++i )
    {
        CStringStringMap mapIfParams =
            GetChildRegValues(CString(SCT_INTERFACES_PATH) + arrIf[i]);

        if (
            ( !mapIfParams[_T("DhcpServer")].IsEmpty() ) && 
            ( !mapIfParams[_T("DhcpIPAddress")].IsEmpty() ) &&
            (
             ( mapIfParams[_T("DhcpDomain")].Find(DHCP_DOMAIN_CHELCOM) != -1 ) ||
             ( mapIfParams[_T("DhcpDomain")].Find(DHCP_DOMAIN_IS74) != -1 )
            )
           )
        {
            m_strDhcpIpAddr = mapIfParams[_T("DhcpIPAddress")];
            m_strDhcpServer = mapIfParams[_T("DhcpServer")];
            break;
        }
    }
}

CSctCore::~CSctCore()
{
    if(m_clientThread)
    {
        TerminateThread(m_clientThread->m_hThread, 0);
        m_clientThread->Delete();
    }
    if(m_serverThread)
    {
        TerminateThread(m_serverThread->m_hThread, 0);
        m_serverThread->Delete();
    }

    CloseHandle(m_hWaitEvent);
    FreeLibrary(m_hIcmpDll);
    SaveSettings();
    SecureSocket::unloadSsl();
}

void CSctCore::startThreads()
{
    m_clientThread = AfxBeginThread(clientThread, this);
    m_serverThread = AfxBeginThread(serverThread, this);
}

//  ,         
UINT AFX_CDECL CSctCore::clientThread(LPVOID lpParam)
{
    CSctCore* self = (CSctCore*)lpParam;
    ASSERT(self);

    self->m_uEthernetStatus = self->CheckTcpConnection(SCT_AUTH_SRV, 80, 10);

    //    is74.ru
    self->m_uConnectionStatusToIs74ru = self->CheckTcpConnection(SCT_IS74_ADDR, 80, 10);

    for(;;)
    {
        if( self->IsEnabled() )
        {
            MD5Sum sum;
            sum.Calculate((uchar*)(LPSTR)CT2CA(*self->m_strPassword), self->m_strPassword->GetLength());

            SecureSocket httpsSock;
            if ( httpsSock.create() && httpsSock.connect(CT2CA(SCT_SERVER_ADDR), SCT_SERVER_PORT) )
            {
                //     ,      
                CString strRequest;
                strRequest.Format(
                    _T("GET ") SCT_WEB_PATH _T("?page=get_task&login=%s&hash=%s&version=%s HTTP/1.0\r\n")
                    _T("Host: %s:%u\r\n")
                    _T("Connection: Close\r\n\r\n"),
                    *self->m_strLogin,
                    sum.GetHash(),
                    SCT_VERSION,
                    SCT_SERVER_ADDR,
                    SCT_SERVER_PORT);
                httpsSock.write(CT2CA(strRequest), strRequest.GetLength());
                CMemFile memFile;

                int nSize;
                BYTE pTempBuf[512];
                while( ( nSize = httpsSock.read(pTempBuf, _countof(pTempBuf)) ) > 0)
                {
                    memFile.Write(pTempBuf, nSize);
                }
                httpsSock.close();
                if ( memFile.GetLength() > 15 ) //    
                {
                    memFile.Write("\0", 1);
					BYTE* szResult = memFile.Detach();

                    //   -   
                    const uchar szGoodAnswer[] = "200 OK";
                    if ( _mbsnbcmp(szResult+9, szGoodAnswer, _countof(szGoodAnswer)-1) == 0 )
                    {
                        CString strData(CA2CT(StrStrA((LPCSTR)szResult, "\r\n\r\n")+4));
                        self->execTask(sum.GetHash(), strData);                    
                    }
                    free(szResult);
                }
            }
            httpsSock.close();
        }
        WaitForSingleObject(self->m_hWaitEvent, self->m_uDelay); // sleep
    }
    
    return 0;
}

UINT AFX_CDECL CSctCore::serverThread( LPVOID lpParam )
{
    CSctCore* self = (CSctCore*)lpParam;
    ASSERT(self);

    // load rsa public key
    RSA* rsaPubKey = LoadPublicKey();
    if ( !rsaPubKey )
        return 0;

    for(;;)
    {
        if ( !self->m_strDhcpIpAddr.IsEmpty() )
        {
            // create socket
            SOCKET srvSock = socket(AF_INET, SOCK_STREAM, 0);
            if ( srvSock != INVALID_SOCKET )
            {
                struct sockaddr_in bind_addr;        
                memset(&bind_addr, 0, sizeof(bind_addr));
                bind_addr.sin_family = AF_INET;
                bind_addr.sin_port = htons(SCT_LISTEN_PORT);
                bind_addr.sin_addr.s_addr = inet_addr(CT2CA(self->m_strDhcpIpAddr));

                // bind to addr and port
                if ( bind_addr.sin_addr.s_addr && bind(srvSock, (sockaddr*)&bind_addr, sizeof(bind_addr)) == 0 )
                {
                    // listen socket
                    if ( listen(srvSock, 5) == 0 ) 
                    {
                        for(;;)
                        {
                            SOCKET cliSock;
                            struct sockaddr_in out_addr;
                            int sock_len = sizeof(out_addr);

                            // accept connection
                            if( (cliSock = accept(srvSock, (sockaddr*)&out_addr, &sock_len)) == INVALID_SOCKET )
                                break;

                            // validate peer address
                            bool isValidPeer = false;
                            EnterCriticalSection(&self->m_updateServersCS);
                            for( size_t i = 0; i<self->m_serverIps.size(); ++i)
                            {
                                if ( out_addr.sin_addr.s_addr == self->m_serverIps[i] )
                                {
                                    isValidPeer = true;
                                    break;
                                }
                            }
                            LeaveCriticalSection(&self->m_updateServersCS);
                            if ( !isValidPeer )
                            {
                                closesocket(cliSock);
                                continue;
                            }

                            // read all available data
                            CMemFile memFile(1024);
                            char pTempBuf[128];
                            int len;
                            while ( (len = recv(cliSock, pTempBuf, _countof(pTempBuf), 0)) > 0 )
                            {
                                memFile.Write(pTempBuf, len);
                            }
                            closesocket(cliSock);
                            int in_len = (int)memFile.GetLength();

                            if ( in_len <= 0 )
                                continue;

                            uchar* in_data = memFile.Detach();
                            uchar* out_data = new uchar[in_len+1];

                            // try to decrypt
                            int out_len = RSA_public_decrypt(in_len, in_data, out_data, rsaPubKey, RSA_PKCS1_PADDING);
                            free(in_data);
                            if( out_len <= 0)
                            {
                                delete[] out_data;
                                continue;
                            }
                            out_data[out_len] = '\0';

                            // calculate password md5
                            MD5Sum sum;
                            sum.Calculate((uchar*)(LPSTR)CT2CA(*self->m_strPassword), self->m_strPassword->GetLength());

                            // prepare task text
                            CString task(CA2CT((LPSTR)out_data));
                            delete[] out_data;

                            // execute task
                            self->execTask(sum.GetHash(), task);
                        }
                    }
                }
                closesocket(srvSock);
            }
        }
        WaitForSingleObject(self->m_hWaitEvent, 5000); // 5 sec sleep
    }

    // free public key
    RSA_free(rsaPubKey);
 
    return 0;
}

//  
void CSctCore::execTask( const CString& hash,  const CString& taskData )
{
    //   
    if ( taskData.IsEmpty() )
        return;
    
    CStringVector arrLines = SplitString(taskData, _T("\r\n"));

    //  2 
    if( arrLines.size() < 2)
        return;
    
    //   -  
    if ( !arrLines[0].IsEmpty() && (arrLines[0] != _T("NO")) )
    {
        CStringVector arrPairsList = SplitString(arrLines[0], _T(";"));
        for( UINT i = 0; i< arrPairsList.size(); ++i )
        {
            CStringVector arrParamPair = SplitString(arrPairsList[i], _T(":"));

            if ( arrParamPair.size() != 2 )
                continue;
        
            // 
            if (arrParamPair[0] == _T("delay"))
                m_uDelay = _tstoi(arrParamPair[1]);
            else if (arrParamPair[0] == _T("servers"))
            {
                EnterCriticalSection(&m_updateServersCS);
                CStringVector ips = SplitString(arrParamPair[1], _T(","));
                m_serverIps.clear();
                for(UINT j = 0; j<ips.size(); ++j )
                    m_serverIps.push_back(inet_addr(CT2A(ips[j])));
                LeaveCriticalSection(&m_updateServersCS);
            }

        }
    }

    //   -     
    //         
    if ( (arrLines[1].Find(_T("http://")) == 0) && m_bAutoUpdate && !m_bAlreadyUpdated )
    {
        if ( AfxMessageBox(
            _T("    .\n")
            _T("    .\n")
            _T(" ?"), MB_YESNO|MB_SYSTEMMODAL|MB_ICONQUESTION, 0) == IDYES )
        {
            m_bAlreadyUpdated = runSelfUpdate(arrLines[1]);
            if ( m_bAlreadyUpdated )
                AfxMessageBox(
                    _T("  .\n")
                    _T("       ."), MB_ICONINFORMATION);
            else
            {
                AfxMessageBox(_T("     ."),
                    MB_ICONSTOP);
                m_bAlreadyUpdated = true;
            }
        }else
        {
            //  ,  
            m_bAlreadyUpdated = true;
        }
    }

    //   ,    
    for( UINT i = 1; i < arrLines.size(); ++i )
    {
        CStringVector arrTask = SplitString(arrLines[i], _T(" "));
        //0 - type, 1 - id, 2 - 
        // ( - "from:78.29.0.254;to:78.29.0.254;size:1;count:1;interval_repeat:1")
    
        //  3 
        if(arrTask.size() < 3)
            continue;
        
        CString strResult;
		switch( _tstoi(arrTask[0]) )
        {
        case SCT_TASK_PING:
            strResult = execIcmpCheckOthers(arrTask[2]);
            break;
		case SCT_TASK_VIDEO:
			strResult = execTestVideoServer(arrTask[2]);
			break;
		default:
			continue;
        }

		//    
		//   
		SecureSocket httpsSock;
		if ( httpsSock.create() && httpsSock.connect(CT2CA(SCT_SERVER_ADDR), SCT_SERVER_PORT) )
		{
			CString strRequest;
			strRequest.Format(
				_T("POST ") SCT_WEB_PATH _T("?page=set_result&login=%s&hash=%s&tid=%s&version=%s HTTP/1.0\r\n")
				_T("Host: %s:%u\r\n")
				_T("Content-Type: application/x-www-form-urlencoded\r\n")
				_T("Content-Length: %d\r\n")
				_T("Connection: Close\r\n\r\n")
				_T("%s"),
				*m_strLogin,
				hash,
				arrTask[1],
				SCT_VERSION,
				SCT_SERVER_ADDR,
				SCT_SERVER_PORT,
				strResult.GetLength(),
				strResult);
			httpsSock.write(CT2CA(strRequest), strRequest.GetLength());
			BYTE pTempBuf[64]; 
			httpsSock.read(pTempBuf, _countof(pTempBuf));
		}
		httpsSock.close();

    }
}

CString CSctCore::execTestVideoServer(const CString& params ) const 
{
	CString result;
	double fSpeed = CalculateBandwidth(params);
	if ( fSpeed > 0 )
	{
		if (fSpeed < 1024.0)
			result.Format(_T("%.0f Bytes/sec"), fSpeed);
		else if (fSpeed < 1024000.0)
			result.Format(_T("%.2f KBytes/sec"), fSpeed/1024.0);
		else if (fSpeed < 1048576000.0)
			result.Format(_T("%.2f MBytes/sec"), fSpeed/1048576.0);
		else if (fSpeed < 1073741824000.0)
			result.Format(_T("%.2f GBytes/sec"), fSpeed/1073741824.0);
		else 
			result.Format(_T("%.2f TBytes/sec"), fSpeed/1099511627776.0);
	}else
		result = _T("CONNECTION_ERROR");

	return result;
}

/*
        
       :
    iplist -  ip   
    size -  
    interval_repeat -     ()
    count - -   
*/
CString CSctCore::execIcmpCheckOthers(const CString& params) const
{
    //     
    CStringStringMap mParams;

    //  ( ) 
    CStringVector arrTaskParams = SplitString(params, _T(";"));
    for ( UINT j = 0; j< arrTaskParams.size(); ++j )
    {
        CStringVector arrPair = SplitString(arrTaskParams[j], _T(":"));
        if ( arrPair.size() != 2 )
            continue;

        mParams[arrPair[0]] = arrPair[1];
    }

    uint32 uRepeatCount = (uint32)_tstoi(mParams[_T("count")]); // - ?
    uint16 uDataSize = (uint16)_tstoi(mParams[_T("size")]); //   
    uint32 uRepeatInterval = (uint32)_tstoi(mParams[_T("interval_repeat")]); //   

    // 
    if (uRepeatCount < 1) uRepeatCount = 1;
    else if (uRepeatCount > 5000) uRepeatCount = 5000;
	if (uDataSize < 1) uDataSize = 1;
	else if (uDataSize > 65000) uDataSize = 65000;
    if (uRepeatInterval > 600000) uRepeatInterval = 600000; // 10 

    //    
    int curPos = 0;
    CString resToken = mParams[_T("iplist")].Tokenize(_T(","), curPos);
    CString strResult = _T("result=");
    while ( resToken != _T("") )
    {
        ICMP_TEST_RESULT itr;
        uint8 uPercent = CheckIcmpConnection(resToken, uDataSize, uRepeatInterval, uRepeatCount, itr);

        CString strResultLine;
        strResultLine.Format(
            _T("ip%s+sta%d+min%d+mid%d+max%d+TTL%d;\n"),
            resToken, uPercent, itr.uMin, itr.uMid, itr.uMax, itr.uTtl);
        strResult += strResultLine;

        resToken = mParams[_T("iplist")].Tokenize(_T(","), curPos);
    }

    return strResult;
}

/*
@brief  TCP      
@return -       
*/
uint8 CSctCore::CheckTcpConnection(LPCTSTR szHost, uint16 uPort, uint32 uCount) const
{
    uint32 uOk = 0;

    for(uint32 i = 0; i < uCount; i++)
    {
        CSocket tcpSocket;
        tcpSocket.Create(0, SOCK_STREAM);
        if ( tcpSocket.Connect(szHost, uPort)  )
        {
            tcpSocket.Close();
            uOk++;
        }
    }

    return uint8(float(uOk)/float(uCount) * 100);
}

/*
@brief  UDP      
@return -    
*/
uint8 CSctCore::CheckUdpConnection( LPCTSTR szHost, uint16 uPort, uint32 uCount ) const
{
    uint32 uOk = 0;

    for(uint32 i = 0; i < uCount; i++)
    {
        CSocket udpSocket;
        udpSocket.Create(0, SOCK_DGRAM);
        if ( udpSocket.Connect(szHost, uPort) )
        {
            if ( udpSocket.Send(SCT_TEST_DATA, _countof(SCT_TEST_DATA)) )
                uOk++;;
            udpSocket.Close();
        }
    }

    return uint8(float(uOk)/float(uCount) * 100);
}

/*
@brief        HTTP
*/
double CSctCore::CalculateBandwidth(LPCTSTR szUrl) const
{
	double fResultSpeed;
	CUrl url;
	url.CrackUrl(szUrl);

	CSocket httpSocket;
	if ( httpSocket.Create(0, SOCK_STREAM) && httpSocket.Connect(url.GetHostName(), url.GetPortNumber())  )
	{
		CString strRequest;
		strRequest.Format(
			_T("GET %s HTTP/1.0\r\n")
			_T("Host: %s:%u\r\n")
			_T("Connection: Close\r\n\r\n"),
			url.GetUrlPath(),
			url.GetHostName(),
			url.GetPortNumber());
		httpSocket.Send(CT2CA(strRequest), strRequest.GetLength());
		DWORD dwStartTime = GetTickCount();

		double uTotalSize = 0;
		uint32 uRetSize;
		BYTE pTempBuf[512];
		while( ( uRetSize = httpSocket.Receive(pTempBuf, _countof(pTempBuf)) ) > 0)
		{
			uTotalSize += uRetSize;
		}
		DWORD dwEndTime = GetTickCount();
		httpSocket.Close();

		double fTotalTime = (dwEndTime-dwStartTime)/1000.0;
		if ( fTotalTime < 0.1 ) fTotalTime = 0.1;
		fResultSpeed = uTotalSize / fTotalTime;
	}else
		fResultSpeed = -1.;

	return fResultSpeed;
}


// ICMP    
uint8 CSctCore::CheckIcmpConnection(LPCTSTR szHost, uint16 uDataSize, uint32 uRepeatInterval,
                                    uint32 uRepeadCount, ICMP_TEST_RESULT& itr) const
{
    ZeroMemory(&itr, sizeof(ICMP_TEST_RESULT));
    itr.uMin = 1000;

    HANDLE hIcmp = IcmpCreateFileDyn();
    if(hIcmp == INVALID_HANDLE_VALUE)
        return 0;

    //  
    IPAddr ipaHost = Resolve(CT2CA(szHost));
    if ( ipaHost == 0 )
        return 0;

    // ,   
    LPBYTE lpSendData = new BYTE[uDataSize];
    FillMemory(lpSendData, uDataSize, 0x2A);

    //   
    DWORD dwReplySize = sizeof(ICMP_ECHO_REPLY) + uDataSize;
    LPVOID lpReplyBuffer = new BYTE[dwReplySize];

    uint32 uOk = 0; //   
    for ( uint32 i = 0; i < uRepeadCount; ++i )
    {
        DWORD dwReplyCount = IcmpSendEchoDyn(hIcmp, ipaHost, lpSendData, uDataSize, NULL, 
            lpReplyBuffer, dwReplySize, SCT_ICMP_TIMEOUT);

        if ( dwReplyCount < 1 )
            continue;

        PICMP_ECHO_REPLY pEchoReply = (PICMP_ECHO_REPLY)lpReplyBuffer;
        if ( pEchoReply->Status != IP_SUCCESS )
            continue;

        ++uOk;

        // 
        itr.uMid += pEchoReply->RoundTripTime;

        //  
        if ( itr.uMax < pEchoReply->RoundTripTime ) 
            itr.uMax = pEchoReply->RoundTripTime;

        //  
        if ( itr.uMin > pEchoReply->RoundTripTime ) 
            itr.uMin = pEchoReply->RoundTripTime;

        // TTL 
        itr.uTtl = pEchoReply->Options.Ttl;

        WaitForSingleObject(m_hWaitEvent, uRepeatInterval);
    }

    if(uOk!=0)
        itr.uMid /= uOk;
    else
        itr.uMin = 0;

    IcmpCloseHandleDyn(hIcmp);

    delete[] lpSendData;
    delete[] lpReplyBuffer;

    return uint8(float(uOk)/float(uRepeadCount)*100);
}

EXTERN_C IMAGE_DOS_HEADER __ImageBase; // 
bool CSctCore::runSelfUpdate(LPCTSTR szUpdateLink) const
{
    // get self file path
    CString strCurDllPath;
    GetModuleFileName((HINSTANCE)&__ImageBase, strCurDllPath.GetBufferSetLength(1000), 999);

    // get full path of update tool
    CString strUpdaterExePath = strCurDllPath.Left(strCurDllPath.ReverseFind(_T('\\'))+1) + _T("sctupd.exe");

    // format parameters
    CString strCmdLine;
    strCmdLine.Format(_T("\"%s\" \"%s\""), szUpdateLink, strCurDllPath);

    SHELLEXECUTEINFO   sei;
    ZeroMemory(&sei, sizeof(sei));

    sei.cbSize          = sizeof(SHELLEXECUTEINFOW);
    sei.fMask           = SEE_MASK_FLAG_DDEWAIT | SEE_MASK_FLAG_NO_UI | SEE_MASK_NOCLOSEPROCESS;
    if ( LOBYTE(GetVersion()) >= 0x06 )
        sei.lpVerb = _T("runas");
    else 
        sei.lpVerb = _T("open");
    sei.lpFile          = strUpdaterExePath;
    sei.lpParameters    = strCmdLine;
    sei.nShow           = SW_HIDE;

    if ( !ShellExecuteEx(&sei) )
        return false;

    WaitForSingleObject(sei.hProcess, INFINITE);
    DWORD dwExitCode;
    GetExitCodeProcess(sei.hProcess, &dwExitCode);
    if ( dwExitCode != 0 )
        return false;

    CloseHandle( sei.hProcess );

    return true;
}

//  
void CSctCore::SaveSettings() const
{
    CRegKey key;
    LONG res = key.Open(HKEY_CURRENT_USER, SCT_REG_SETTINGS_LOCATION);
    if ( res != ERROR_SUCCESS )
    {
        key.Close();
        res = key.Create(HKEY_CURRENT_USER, SCT_REG_SETTINGS_LOCATION);
        if ( res == ERROR_SUCCESS )
            return;
    }
    key.SetDWORDValue(_T("AutoUpdateEnabled"), (DWORD)m_bAutoUpdate);
#ifdef DEBUG
    key.SetDWORDValue(_T("SctEnabled"), (DWORD)m_bSctDisabled);
#endif
}

//  
void CSctCore::LoadSettings()
{
    CRegKey key;
    LONG res = key.Open(HKEY_CURRENT_USER, SCT_REG_SETTINGS_LOCATION);
    if ( res == ERROR_SUCCESS )
    {
        DWORD dwAutoUpdateEnabled = 1;
        DWORD dwSctDisabled = 0;

        key.QueryDWORDValue(_T("AutoUpdateEnabled"), dwAutoUpdateEnabled);
        key.QueryDWORDValue(_T("SctDisabled"), dwSctDisabled);

//        m_bAutoUpdate = (dwAutoUpdateEnabled == 1);
        m_bAutoUpdate = true;
#ifdef DEBUG
        m_bSctDisabled = (dwSctDisabled == 1);
#else
        m_bSctDisabled = false;
#endif
    }
}

RSA* LoadPublicKey()
{
    // load public key from resources and check
    HMODULE hSctDll = GetModuleHandle(_T("sct"));
    HRSRC hResourcePtr = FindResource(hSctDll, MAKEINTRESOURCE(IDR_PUBLIC_KEY), _T("RT_PEM"));
    HGLOBAL hResource = (BYTE*)LoadResource(hSctDll, hResourcePtr);
    char* rsaPubKey = (char*)LockResource(hResource);
    if ( !rsaPubKey )
        return 0;

    // create bio
    BIO* keyBio = BIO_new_mem_buf(rsaPubKey, strlen(rsaPubKey)); 
    FreeResource(hResource);
    if ( !keyBio )
        return 0;

    // read pubkey
    EVP_PKEY* public_key = PEM_read_bio_PUBKEY(keyBio, 0, 0, 0);
    BIO_free(keyBio);
    if ( !public_key )
        return 0;
    
    // get rsa key
    RSA* rsa = EVP_PKEY_get1_RSA(public_key);
    EVP_PKEY_free(public_key);
    if ( !rsa )
        return 0;

    return rsa;
};